#!/usr/local/bin/perl -w
#
# $Id: rtimportldap.pl,v 1.20 2003/01/13 09:34:56 stsiny Exp $
# RT is (c) 1996-2001 Jesse Vincent <jesse@fsck.com>

# rtimportldap.pl: the utility for importing the RT users from LDAP directory
# Author: Stanislav Sinyagin <ssinyagin@yahoo.com>
# Some parts of the code are inherited from RT's rtadmin
# LDAP group mapping is contributed by
#  Andreas (andi) Hofmeister <andreas.hofmeister@pyramid.de>

# This program is distributed under GNU GPL.
# Copyright (c) Stanislav Sinyagin 2002

# ASSUMPTIONS:
#   This script does not care about passwords.
#
# TESTED:
#   + in Microsoft Active Directory/MS Exchange 2000 environment,
#     with users being authenticated by Apache/mod_auth_ldap against AD
#
#   + OpenLDAP with inetOrgPerson + NIS schema, users
#     are authenticated using pam_ldap (andi)
#
# NOTES:
#   Some fields are defined in RFC1274 (cosine.schema), but no
#   object class contains them. MS AD contains them, supposedly in
#   object class 'user'. 
#
#   'company' is used in MS AD, but not defined in standard schemas. We map
#   it to 'Organization'.
#
#   RT uses HTML::Mason, which uses HTML::Entity for displaying
#   non-ASCII characters. As described in HTML::Entity, it conciders
#   the input string as Unicode if Perl version is more than 5.7,
#   and ISO-8859-1 (Latin1) otherwise.
#   --latin1 option converts the UTF-8 values of LDAP entries into
#   Latin1 before passing the data to RT. 
#   This is tested with Perl 5.6.1 and RT 2.0.15pre1 only.
#
#   MS AD uses proprietary schema for group mapping. The object class is
#   "group", and the member attribute "member" contains DNs of the users,
#   instead of UIDs. This script understands both NIS or MS AD ways.

##########################################################################
#  Libraries used

use strict;
use Net::LDAP;
use Getopt::Long;
use Unicode::String;
    
# Replace this with your RT_LIB_PATH
use lib "/usr/local/rt2/lib";
# Replace this with your RT_ETC_PATH
use lib "/usr/local/rt2/etc";

use RT::Interface::CLI  qw(CleanEnv LoadConfig DBConnect 
                           GetCurrentUser GetMessageContent);
use RT::Group;

##########################################################################
# RT CLI initialization
 
#Clean out all the nasties from the environment
CleanEnv();

#Load etc/config.pm and drop privs
LoadConfig();

#Connect to the database and get RT::SystemUser and RT::Nobody loaded
DBConnect();

#Drop setgid permissions
RT::DropSetGIDPermissions();

#Get the current user all loaded
our $CurrentUser = GetCurrentUser();

unless( $CurrentUser->Id )
{
    print "No RT user found. Please consult your RT administrator.\n";   
    exit 1;
}

##########################################################################
# Process options

our %opts;
my %cliOpts;

$opts{'port'}          = 389;
$opts{'ldapver'}       = 3;
$opts{'filter'}        =
    '(|(objectClass=organizationalPerson)(objectClass=person))';
$opts{'uidattr'}       = 'uid';
$opts{'attrmap'}       = [];
$opts{'disabled'}      = 0;
$opts{'privileged'}    = 1;
$opts{'groupadd'}      = [];
$opts{'groupsync'}     = [];
$opts{'groupmap'}      = [];
$opts{'gmfilter'}      = '(objectClass=posixGroup)';
$opts{'gmgidattr'}     = 'cn';
$opts{'gmmemattr'}     = 'memberUid';
$opts{'gmdescattr'}    = 'description';
# see Notes
$opts{'latin1'} = $] > 5.007 ? 0:1;

my $ok = GetOptions(\%cliOpts,
                    'rc=s',
                    'server=s',
                    'port=s',
                    'ldapver=i',
                    'binddn=s',
                    'bindpw=s',
                    'basedn=s',
                    'filter=s',
                    'uidattr=s',
                    'attrmap=s@',
                    'groupadd=s@',
                    'groupsync=s@',
                    'groupmap=s@',
                    'gmbase=s',
                    'gmfilter=s',
                    'gmgidattr=s',
                    'gmmemattr=s',
                    'disabled=i',
                    'privileged=i',
                    'latin1=i',
                    'verbose',
                    'help|h');

if( $cliOpts{'rc'} )
{
    # Read options from RC file and then override them with CLI options
    
    open( RC, $cliOpts{'rc'} ) or
        die("Error opening " . $cliOpts{'rc'} . ": $!\n");
    while( <RC> )
    {
        # Pass comments and empty lines
        next if /^#/;
        next if /^\s*$/;
        if( /^(\w+)\=(\S.*\S)\s*$/ )
        {
            if( ref( $opts{$1} ) )
            {
                push( @{$opts{$1}}, $2 );
            }
            else
            {
                $opts{$1} = $2;
            }
        }
        else
        {
            printf STDERR ("Invalid option format in %s: %s\n",
                           $cliOpts{'rc'}, $_);
            exit 1;
        }
    }
    close RC;                           
}

while( my ($opt, $val) = each %cliOpts )
{
    if( ref( $opts{$opt} ) )
    {
        push( @{$opts{$opt}}, @{$val} );
    }
    else
    {        
        $opts{$opt} = $val;
    }
}

if( not $ok or scalar(@ARGV) > 0 or 
    $opts{'help'} or
    not $opts{'server'} or
    not  $opts{'basedn'} )
{
    print STDERR
        "\n" .
        "Script for importing LDAP users into RT: Request Tracker\n" .
        "User entries found by the LDAP search filter are exported to RT\n" .
        "database. In addition, they may be added to RT groups, according\n" .
        "to --groupadd, --groupsync, and --groupmap options.\n" .
        "Missing RT groups are created automatically.\n" .
        "\n";
    print STDERR
        "Usage: $0 options...\n" .
        "Options ([M] means mandatory, [O] means optional, " .
        "+ means multivalue):\n" .
        "  --rc         filename\n" .
        "      [O]  read options from file: opt=value pairs one per line\n" .
        "  --server     hostname\n" . 
        "      [M]  LDAP host name or address\n" .
        "  --port       port\n" .
        "      [O]  LDAP TCP port. Default: " . $opts{'port'} . "\n" .
        "  --ldapver    3|2\n" .
        "      [O]  LDAP version. Default: " . $opts{'ldapver'} . "\n" .
        "  --binddn     dn\n" .
        "      [O]  Bind DN\n" .
        "  --bindpw     password\n" .
        "      [O]  Bind password\n" .
        "  --basedn     dn\n" .
        "      [M]  Base DN\n" .
        "  --filter     filter\n" .
        "      [O]  LDAP search filter.\n" .
        "           Default: " . $opts{'filter'} . "\n" .
        "  --uidattr    attr\n" .
        "      [O]  Username attribute. Default: " . $opts{'uidattr'} . "\n" .
        "  --attrmap    lattr:rattr\n" .
        "      [O+] Mapping of LDAP to RT attribute. Use lattr:none to\n" .
        "           suppress standard mapping of lattr\n" .
        "  --groupadd   groupname\n" .
        "      [O+] Add all found LDAP users to these RT groups\n" .
        "  --groupsync  groupname\n" .
        "      [O+] Maintain exact matching of the LDAP search result and\n" .
        "           the RT group membership\n" .
        "  --groupmap   ldapgrp[:rtgrp]\n" .
        "      [O+] Synchronize members of LDAP group with members\n" .
        "           of RT group. LDAP group lookup is controlled by\n" .
        "           --gmbase, --gmfilter, --gmgidattr, and --gmmemattr\n" .
        "           options. Several LDAP groups may be merged into the\n" .
        "           same RT group. Group members must be found by --filter\n" .
        "           before group mapping.\n" .
        "  --gmbase     dn\n" .
        "      [O]  LDAP group search DN. Default: same as --basedn\n" .
        "  --gmfilter   filter\n" .
        "      [O]  Group mapping LDAP filter. Default: " .
        $opts{'gmfilter'} . "\n" .
        "  --gmgidattr  attr\n" .
        "      [O]  Group mapping GID attribute. Default: " .
        $opts{'gmgidattr'} . "\n" .
        "  --gmmemattr  attr\n" .
        "      [O]  Group mapping member attribute. Default: " .
        $opts{'gmmemattr'} . "\n" .
        "  --gmdescattr attr\n" .
        "      [O]  Group mapping description attribute. Default: " .
        $opts{'gmdescattr'} . "\n" .        
        "  --disabled   1|0\n" .
        "      [O]  Set RT users as disabled. Default: " .
        $opts{'disabled'} . "\n" .
        "  --privileged 1|0\n" .
        "      [O]  Set RT users as privileged. Default: " .
        $opts{'privileged'} . "\n" .
        "  --latin1     1|0\n" .
        "      [O]  Convert LDAP values to Latin1 encoding." .
        " Default: " . $opts{'latin1'} . "\n" .
        "  --verbose\n" .
        "      [O]  Print extra information\n" .
        "  --help\n" .
        "      [O]  This help message\n" .
        "\n";
    exit 1;
}

if( not $opts{'gmbase'} )
{
    $opts{'gmbase'} = $opts{'basedn'};
}

##########################################################################
# Load RT groups

my %groupadd_hash = ();
if( @{$opts{'groupadd'}} > 0 )
{
    loadRTGroups( \%groupadd_hash, @{$opts{'groupadd'}} );
}

my %groupsync_hash = ();
if( @{$opts{'groupsync'}} > 0 )
{
    loadRTGroups( \%groupsync_hash, @{$opts{'groupsync'}} );
}

my %groupmap_hash = ();
if( @{$opts{'groupmap'}} > 0 )
{
    foreach my $gmappair ( @{$opts{'groupmap'}} )
    {
        my( $ldapGroup , $rtGroup ) = split( ":", $gmappair );
        
        if( not $rtGroup )
        {
            $rtGroup = $ldapGroup;
        }
        
        # If several LDAP groups are merging, this is initialized once
        if( not exists( $groupmap_hash{$rtGroup} ) )
        {
            $groupmap_hash{$rtGroup}{'ldapGroups'} = [];
        }

        push( @{$groupmap_hash{$rtGroup}{'ldapGroups'}}, $ldapGroup );
    }
    loadRTGroups( \%groupmap_hash, keys %groupmap_hash );
}

##########################################################################
# Initialize the user attributes mapping

our %ldap_user_attr_map = ( 'mail'                  => 'EmailAddress',
                            'info'                  => 'FreeformContactInfo',
                            'description'           => 'Comments',
                            # 'company' is the Microsoft way to say things
                            'company'               => 'Organization',
                            'o'                     => 'Organization',
                            'organizationName'      => 'Organization',
                            'cn'                    => 'RealName',
                            'homePhone'             => 'HomePhone',
                            'homeTelephoneNumber'   => 'HomePhone',
                            'telephoneNumber'       => 'WorkPhone',
                            'mobile'                => 'MobilePhone',
                            'mobileTelephoneNumber' => 'MobilePhone',
                            'pager'                 => 'PagerPhone',
                            'pagerTelephoneNumber'  => 'PagerPhone',
                            # Where to put physicalDeliveryOfficeName ?
                            'street'                => 'Address1',
                            'streetAddress'         => 'Address1',
                            'l'                     => 'City',
                            'localityName'          => 'City',
                            'st'                    => 'State',
                            'stateOrProvinceName'   => 'State',
                            'postalCode'            => 'Zip',
                            'c'                     => 'Country',
                            'countryName'           => 'Country',
                            'gecos'                 => 'Gecos'
                            );

# Process the custom attribute mapping
if( exists( $opts{'attrmap'} ) )
{
    foreach my $pair ( @{$opts{'attrmap'}} )
    {
        my ($lattr, $rattr) = split(':', $pair);
        if( $lattr =~ /\w/ and $rattr =~ /\w/ )
        {
            $ldap_user_attr_map{$lattr} = $rattr;
        }
        else
        {
            printf STDERR ("Incorrect mapping: %s\n", $pair);
            exit 1;
        }
    }
}

# UID mapping
$ldap_user_attr_map{$opts{'uidattr'}} = 'Name';

# All LDAP user attributes of our interest
our @ldap_attr_list = keys %ldap_user_attr_map;


##########################################################################
# Initialize LDAP connection

# Make sure that UID mapping is set correctly
$opts{'filter'} = sprintf( '(&(%s=*)%s)', $opts{'uidattr'}, $opts{'filter'} );

# Bind to LDAP directory
my $ldap = Net::LDAP->new( $opts{'server'},
                           'port'    => $opts{'port'},
                           'onerror' => 'die',
                           'version' => $opts{'ldapver'} );
if( not defined( $ldap ) )
{
    printf STDERR ("Error connecting to %s: %s\n",
                   $opts{'server'}, $@);
    exit 1;
}

if( not $opts{'binddn'} )
{
    $ldap->bind();
}
else
{
    $ldap->bind( $opts{'binddn'}, 'password' => $opts{'bindpw'} );
}

if( $opts{'verbose'} )
{
    printf("Connected to %s:%s\n", $opts{'server'}, $opts{'port'});
}

##########################################################################
# Search for LDAP users

if( $opts{'verbose'} )
{
    printf( "Sending LDAP search: base=\"%s\", filter=\"%s\"\n",
            $opts{'basedn'}, $opts{'filter'} );
}

my $mesg = $ldap->search( 'base'   => $opts{'basedn'},
                          'scope'  => 'sub',
                          'filter' => $opts{'filter'},
                          'attrs'  => \@ldap_attr_list);
if( $mesg->code() )
{
    printf STDERR ("LDAP error: %s\n", $mesg->error());
    exit 1;
}

my @userEntries = $mesg->entries();

if( @userEntries == 0 )
{
    print STDERR ("ERROR: LDAP user search returned no entries\n");
    exit 1;
}

if( $opts{'verbose'} )
{
    printf( "Found %d LDAP user entries\n", scalar( @userEntries ) );
}

##########################################################################
# Search for LDAP groups

foreach my $gname ( sort keys %groupmap_hash )
{
    foreach my $ldapGroup ( sort @{$groupmap_hash{$gname}{'ldapGroups'}} )
    {
        my $gfilter = sprintf( "(&%s(%s=%s))", 
                               $opts{'gmfilter'} , 
                               $opts{'gmgidattr'} ,
                               $ldapGroup );

        if( $opts{'verbose'} )
        {
            printf( "Sending LDAP search: base=\"%s\", filter=\"%s\"\n",
                    $opts{'gmbase'}, $gfilter );
        }

        my $mesg = $ldap->search( 'base'   => $opts{'gmbase'},
                                  'scope'  => 'sub',
                                  'filter' => $gfilter,
                                  'attrs'  => [ $opts{'gmmemattr'},
                                                $opts{'gmdescattr'} ] );
    
        if( $mesg->code() )
        {
            printf STDERR ("LDAP error: %s\n", $mesg->error());
            exit 1;
        }
        
        my @groupEntries = $mesg->entries();

        if( scalar( @groupEntries ) != 1 )
        {
            printf STDERR ("Expected one LDAP entry, but found %d ".
                           "entries for group %s\n", 
                           scalar( @groupEntries ), $ldapGroup );
            exit 1;
        }
        
        my $entry = $groupEntries[0];
        my @values = $entry->get_value( $opts{'gmmemattr'} );
      
        if( $opts{'verbose'} )
        {
            printf( "Found %d members in LDAP group %s \n" , 
                    scalar( @values ), $ldapGroup );
        }

        foreach my $member ( @values )
        {
            $groupmap_hash{$gname}{'ldapMembers'}{$member} = 1;
        }

        # Remember the group description
    
        my $descr = $entry->get_value( $opts{'gmdescattr'} );
        if( defined( $descr ) )
        {
            if( $opts{'latin1'} )
            {
                my $u = Unicode::String::utf8( $descr );
                $descr = $u->latin1();
            }
            $groupmap_hash{$gname}{'ldapDescr'}  = $descr;
        }
    }
}

$ldap->unbind();


##########################################################################
# Update RT users

my %ldap_users_dn = ();

$ok = 1;

foreach my $entry ( @userEntries )
{
    $ok = updateRTUser( $entry, \%ldap_users_dn ) ? $ok:0;
}

if( not $ok )
{
    print STDERR ("Cannot perform group processing because of some errors\n");
    exit 1;
}


##########################################################################
# Group processing.
# We assume that all group members are already valid RT users.

my @ldap_users = values %ldap_users_dn;

addGroupMembers( \%groupadd_hash, @ldap_users );
syncGroupMembers( \%groupsync_hash, @ldap_users );

foreach my $gname ( sort keys %groupmap_hash )
{
    my $group_obj = $groupmap_hash{$gname}{'rtGroupObj'};

    if( exists( $groupmap_hash{$gname}{'ldapDescr'} ) and
        $group_obj->Description() ne $groupmap_hash{$gname}{'ldapDescr'} )
    {
        my ($val, $msg) =
            $group_obj->SetDescription( $groupmap_hash{$gname}{'ldapDescr'} );
        if( $val )
        {
            printf("Updated Description for RT group %s\n", $gname);
        }
        else
        {
            printf("ERROR updating Description for RT group %s: %s\n",
                   $gname, $msg);
        }
    }

    my @members = ();
    foreach my $member ( keys %{$groupmap_hash{$gname}{'ldapMembers'}} )
    {
        # in MS AD, the user DN is put into member list
        if( $member =~ /=/ )
        {
            if( $ldap_users_dn{$member} )
            {
                $member = $ldap_users_dn{$member};
            }
            else
            {
                printf STDERR ("ERROR: Unknown member DN: %s\n", $member);
                exit 1;
            }
        }
        push( @members, $member );
    }
            
    syncGroupMembers( \%groupmap_hash, @members );
}

exit 0;

##########################################################################
###                          SUBROUTINES                               ###
##########################################################################


##########################################################################
# $ok = updateRTUser( $entry, \%userDnList );
# %userDnList is filled with $dn => $uname pairs

# Process the LDAP entry and create/update the RT user

sub updateRTUser
{
    my $entry = shift;
    my $dnList = shift;

    my %userArgs =  ( 'Privileged' => $opts{'privileged'},
                      'Disabled'   => $opts{'disabled'} );
    
    my %updatedBy = ( 'Privileged' => 'thisScript',
                      'Disabled'   => 'thisScript' );
    
    foreach my $lattr ( @ldap_attr_list )
    {
        if( $entry->exists( $lattr ) )
        {
            my $value = $entry->get_value( $lattr );
            my $rattr = $ldap_user_attr_map{$lattr};
            if( $rattr ne 'none' )
            {
                if( exists( $userArgs{$rattr} ) )
                {
                    printf( "WARNING: Attributes %s and %s conflict mapping ".
                            "to %s in DN=%s\n", $lattr,
                            $updatedBy{$rattr}, $rattr, $entry->dn() );
                }
                else
                {
                    if( $opts{'latin1'} )
                    {
                        my $u = Unicode::String::utf8( $value );
                        $value = $u->latin1();
                    }
                    $userArgs{$rattr} = $value;
                    $updatedBy{$rattr} = $lattr;
                }
            }
        }
    }

    # Now we have all user attributes in %userArgs

    my $ok = 1;
    
    my $user_obj = new RT::User($CurrentUser);
    my $uname = $userArgs{'Name'};
    
    # Check if the user exists
    $user_obj->Load( $uname );    
    if( not $user_obj->id() )
    {
        my @r = $user_obj->Create( %userArgs );
        if( not $r[0] )
        {
            printf("ERRORR creating RT user %s: %s\n", $uname, $r[1]);
            $ok = 0;
        }
        printf( "New RT user created: %s (%s)\n",
                $uname, $userArgs{'RealName'} );        
    }
    else
    {
        if( $opts{'verbose'} )
        {
            printf("Checking up user: %s (%s)\n",
                   $uname, $userArgs{'RealName'});
        }
        foreach my $rattr ( keys %userArgs )
        {
            if( $user_obj->$rattr() ne $userArgs{$rattr} )
            {
                my $method = 'Set'.$rattr;
                my ($val, $msg) = $user_obj->$method( $userArgs{$rattr} );
                if( $val )
                {
                    printf("Updated %s for %s (%s)\n",
                           $rattr, $uname, $userArgs{'RealName'});
                }
                else
                {
                    printf("ERROR updating %s for %s (%s): %s\n",
                           $rattr, $uname, $userArgs{'RealName'}, $msg);
                    $ok = 0;
                }
            }
        }                   
    }
    
    if( $ok )
    {
        if( ref( $dnList ) )
        {
            $dnList->{$entry->dn()} = $uname;
        }
    }
    
    return $ok;
}

    
##########################################################################
# loadRTGroups( \%hash, @groupNames )

# Kind-of OO approach. We store the group parameters in
# a hash: { groupName => { 'rtGroupObj' => {RT::Group object},
#                          'rtMembers'  => [],
#                          'rtMembersNotInLdap' => { uname => 1 } } }
# 'rtMembers' is not yet used anywhere

sub loadRTGroups
{
    my $gHash = shift;
    my @groupNames = @_;
    
    foreach my $gname ( @groupNames )
    {
        my $group_obj = new RT::Group( $CurrentUser );
        $group_obj->Load( $gname );
        
        if( not $group_obj->Id() )
        {
            if( $opts{'verbose'} )
            {
                printf( "RT group not found: %s\n", $gname );
            }
            $group_obj->Create( 'Name' => $gname );
            if( not $group_obj->Id() )
            {
                printf STDERR ( "Could not create group: %s\n", $gname );
                exit 1;
            }
            printf( "Created RT group: %s\n", $gname );            
        }
        
        $gHash->{$gname}->{'rtGroupObj'} = $group_obj;

        $gHash->{$gname}->{'rtMembers'} = [];
        # Make sure the hash always exists, even if empty
        $gHash->{$gname}->{'rtMembersNotInLdap'} = {};
        
        my $members = $group_obj->MembersObj();
        while( my $member = $members->Next() )
        {
            my $user_obj = $member->UserObj();
            my $uname = $user_obj->Name();
            push( @{$gHash->{$gname}->{'rtMembers'}}, $uname );
            $gHash->{$gname}->{'rtMembersNotInLdap'}->{$uname} = 1;
        }
    }
}

##########################################################################
# addGroupMembers( \%hash, @userList )
# Add members to RT group

sub addGroupMembers
{
    my $gHash = shift;
    my @uList = @_;

    foreach my $gname ( keys %{$gHash} )
    {
        my $group_obj = $gHash->{$gname}->{'rtGroupObj'};

        foreach my $uname ( sort @uList )
        {
            my $user_obj = new RT::User( $CurrentUser );
            $user_obj->Load( $uname );    
            if( not $user_obj->id() )
            {
                printf STDERR ("ERROR: User %s is supposed to be in RT " .
                               "database before adding to group %s\n",
                               $uname, $gname);
                exit 1;
            }
            
            if( not defined( $group_obj->HasMember( $user_obj->id() ) ) )
            {
                my ($val, $msg) = $group_obj->AddMember( $uname );
                if( $val )
                {
                    printf( "Added %s to group %s\n", $uname, $gname );
                }
                else
                {
                    printf( "ERROR adding %s to group %s: %s\n",
                            $uname, $gname, $msg );
                }
            }

            delete $gHash->{$gname}->{'rtMembersNotInLdap'}->{$uname};
        }
    }
}


##########################################################################
# syncGroupMembers( \%hash, @userList )
# Synchronize member list with RT group

sub syncGroupMembers
{
    my $gHash = shift;
    my @uList = @_;

    addGroupMembers( $gHash, @uList );

    # Now delete from RT groups those users not found in LDAP

    foreach my $gname ( keys %{$gHash} )
    {
        my $group_obj = $gHash->{$gname}->{'rtGroupObj'};

        foreach my $uname
            ( sort keys %{$gHash->{$gname}->{'rtMembersNotInLdap'}} )
        {
            my ($val, $msg) =  $group_obj->DeleteMember( $uname );
            if( $val )
            {
                printf("Deleted %s from group %s\n", $uname, $gname);
            }
            else
            {
                printf("ERROR deleting %s from group %s: %s\n",
                       $uname, $gname, $msg);
            }            
        }
    }
}    

# Whitespace police

# Local Variables:
# mode: perl
# indent-tabs-mode: nil
# perl-indent-level: 4
# End:
